#include <avr/io.h>
#include <avr/portpins.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <elapsedMillis.h>
#include <errno.h>

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

/**

ALARM <- PB5
IGNITION <- PB4
TACHO <- PD6=T1 or PD0=INT0
KILL -> PD4
CONNECTED <- PE2
LED -> PF0
ENABLE <- PF4
OVERRIDE <- PF6

**********************************/

#define BUFLEN 30
volatile uint32_t rpm;
elapsedMicros tacho;
elapsedMillis reporttimer, ledtimer;
uint8_t override, ignition, enable, kill, force, connected, echo, alarm, noauto, usbconnected;
char usbbuf[BUFLEN];

ISR(INT0_vect) {
    rpm = 30000000UL / tacho;
    tacho = 0;
}

void bootloader() {
    wdt_disable();
    cli();
    UDCON = 1;
    USBCON = (1<<FRZCLK);  // disable USB
    UCSR1B = 0;
    delay(5);
    EIMSK = 0; PCICR = 0; SPCR = 0; ACSR = 0; EECR = 0; ADCSRA = 0;
    TIMSK0 = 0; TIMSK1 = 0; TIMSK3 = 0; TIMSK4 = 0; UCSR1B = 0; TWCR = 0;
    DDRB = 0; DDRC = 0; DDRD = 0; DDRE = 0; DDRF = 0; TWCR = 0;
    PORTB = 0; PORTC = 0; PORTD = 0; PORTE = 0; PORTF = 0;
    asm volatile("jmp 0x7000");
}

void setup() {
    DDRB = DDRC = DDRD = DDRE = 0;
    PORTB = PORTC = PORTD = PORTE = 0;
    Serial.begin(4800);
    EICRA |= (1<<ISC01);                        // Falling edge of INT0 generates interrupt
    EIMSK |= (1<<INT0);                         // Enable INT0

    PORTF |= (1<<PF6);                          // Pullup on override pin
    DDRD |= (1<<PD4);                           // Kill signal to output
    DDRD |= (1<<PF0);                           // LED to output
    wdt_enable(WDTO_2S);
}

void loop() {
    static uint8_t bufoff;

    wdt_reset();
    if (Serial && !usbconnected) {
         usbconnected = true;
    } else if (!Serial && usbconnected) {
        usbconnected = false;
        echo = noauto = 0;
        bufoff = 0;
    }
    if (Serial && Serial.available()) {
        int c = Serial.read();
        if (c == 10 || c == 13 || c == 0 || bufoff == BUFLEN - 1) {
            usbbuf[bufoff] = 0;
            if (echo) {
                Serial.println();
            }
            if (bufoff > 0) {
                if (!strcmp(usbbuf, "help")) {
                    Serial.print("Ignition Control");
                    Serial.println(F("help              this message"));
                    Serial.println(F("bootloader        jump to bootloader"));
                    Serial.println(F("kill              kill engine"));
                    Serial.println(F("nokill            unkill engine"));
                } else if (!strcmp(usbbuf, "bootloader")) {
                    bootloader();
                } else if (!strcmp(usbbuf, "auto")) {
                    noauto = 0;
                } else if (!strcmp(usbbuf, "noauto")) {
                    noauto = 1;
                } else if (!strcmp(usbbuf, "kill")) {
                    kill = 1;
                    PORTD |= (1<<PD4);
                } else if (!strcmp(usbbuf, "nokill")) {
                    kill = 0;
                    PORTD &= ~(1<<PD4);
                } else {
                    Serial.print("Unknown command \"");
                    Serial.print(usbbuf);
                    Serial.println('"');
                }
            }
            bufoff = 0;
        } else {
            usbbuf[bufoff++] = c;
            if (echo) {
                Serial.print((char)c);
            }
        }
    }

    uint8_t newconnected = PINE & (1<<PE2) ? 1 : 0;
    uint8_t newoverride  = PINF & (1<<PF6) ? 0 : 1;
    uint8_t newenable    = PINF & (1<<PF4) ? 1 : 0;
    uint8_t newignition  = PINB & (1<<PB4) ? 1 : 0;
    uint8_t newalarm     = newenable && (PINB & (1<<PB5)) ? 1 : 0;
    if (connected != newconnected) {
        connected = newconnected;
        force = 5;
    }
    if (override != newoverride) {
        override = newoverride;
        force = 1;
    }
    if (alarm != newalarm) {
        alarm = newalarm;
        force = 2;
    }
    if (enable != newenable) {
        enable = newenable;
        force = 3;
    }
    if (ignition != newignition) {
        ignition = newignition;
        force = 4;
    }
    if (tacho > 1000000) {
        rpm = 0;
        tacho = 0;
    }

    if ((!noauto && reporttimer > 1000) || force) {
        Serial.print("{\"type\":\"ignition\",\"rev\":\"" STR(FIRMWAREREVISION) "/" STR(HARDWAREREVISION) "\",\"connected\":");
        Serial.print(connected?"true":"false");
        Serial.print(",\"on\":");
        Serial.print(enable?"true":"false");
        Serial.print(",\"kill\":");
        Serial.print(kill?"true":"false");
        Serial.print(",\"alarm\":");
        Serial.print(alarm?"true":"false");
        Serial.print(",\"rpm\":");
        Serial.print(rpm);
        Serial.print(",\"ignition\":");
        Serial.print(ignition?"true":"false");
        Serial.print(",\"override\":");
        Serial.print(override?"true":"false");
        Serial.println("}");
        reporttimer = 0;
        force = 0;
    }

    // Alarm - rapid flash
    // Kill - brief flash once per second
    // Override - 2s isophase

    if (alarm) {
        if (ledtimer >= 200) {
            ledtimer = 0;
        }
        if (ledtimer < 100) {
            PORTF |= (1<<PF0);
        } else {
            PORTF &= ~(1<<PF0);
        }
    } else if (kill) {
        if (ledtimer >= 1000) {
            ledtimer = 0;
        }
        if (ledtimer < 100) {
            PORTF |= (1<<PF0);
        } else {
            PORTF &= ~(1<<PF0);
        }
    } else if (override) {
        if (ledtimer >= 4000) {
            ledtimer = 0;
        }
        if (ledtimer < 2000) {
            PORTF |= (1<<PF0);
        } else {
            PORTF &= ~(1<<PF0);
        }
    } else {
        PORTF |= (1<<PF0);
    }
}
